from base64 import b64encode as encode
import hashlib
import inspect
import os
from typing import Any, Dict, List, Optional, Sequence, Union, cast

from django.core.exceptions import ValidationError, FieldDoesNotExist
try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_str as force_text  # type: ignore
from django.db.models.signals import (
    class_prepared,
    post_init,
    pre_init
)
from ldaporm.fields import Field
from ldaporm.managers import LdapManager
from ldaporm.typing import LDAPData

from .options import Options


class LdapModelBase(type):

    def __new__(cls, name, bases, attrs, **kwargs):
        super_new = super().__new__
        parents = [b for b in bases if isinstance(b, LdapModelBase)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        # Create the class.
        module = attrs.pop('__module__')
        new_attrs = {'__module__': module}
        classcell = attrs.pop('__classcell__', None)
        if classcell is not None:
            new_attrs['__classcell__'] = classcell
        new_class = super_new(cls, name, bases, new_attrs, **kwargs)
        attr_meta = attrs.pop('Meta', None)
        meta = attr_meta or getattr(new_class, 'Meta', None)

        # Add our Meta class.  This simluates the Django ORM Meta class
        # enough that ModelForm will work for us, among other things
        new_class.add_to_class('_meta', Options(meta))

        # Add all attributes to the class.  This is where the fields get
        # initialized
        for obj_name, obj in attrs.items():
            new_class.add_to_class(obj_name, obj)

        new_class._meta.concrete_model = new_class
        new_class._prepare()

        return new_class

    def add_to_class(cls, name: str, value: Any) -> None:
        # We should call the contribute_to_class method only if it's bound
        if not inspect.isclass(value) and hasattr(value, 'contribute_to_class'):
            value.contribute_to_class(cls, name)
        else:
            setattr(cls, name, value)

    def _prepare(cls) -> None:
        """
        Create some methods once self._meta has been populated.

        Importantly, this is where the Manager class gets added.
        """
        opts = cls._meta   # type: ignore
        opts._prepare(cls)

        # Give the class a docstring -- its definition.
        if cls.__doc__ is None:
            cls.__doc__ = "%s(%s)" % (cls.__name__, ", ".join(f.name for f in opts.fields))

        if any(f.name == 'objects' for f in opts.fields):
            raise ValueError(
                "Model %s must specify a custom Manager, because it has a "
                "field named 'objects'." % cls.__name__
            )
        manager = opts.manager_class()
        cls.add_to_class('objects', manager)  # pylint: disable=no-value-for-parameter
        class_prepared.send(sender=cls)


class Model(metaclass=LdapModelBase):

    class DoesNotExist(Exception):
        pass

    class InvalidField(Exception):
        pass

    class MultipleObjectsReturned(Exception):
        pass

    _meta: Optional[Options] = None
    objects: Optional[LdapManager] = None

    def __init__(self, *args, **kwargs) -> None:
        cls = self.__class__
        opts = cast(Options, self._meta)
        _setattr = setattr
        self._dn: Optional[str] = None

        pre_init.send(sender=cls, args=args, kwargs=kwargs)

        if len(args) > len(opts.fields):
            # Daft, but matches old exception sans the err msg.
            raise IndexError("Number of args exceeds number of fields")

        if not kwargs:
            fields_iter = iter(opts.fields)
            for val, field in zip(args, fields_iter):
                _setattr(self, cast(str, field.name), val)
        else:
            # Slower, kwargs-ready version.
            fields_iter = iter(opts.fields)
            for val, field in zip(args, fields_iter):
                _setattr(self, cast(str, field.name), val)
                kwargs.pop(cast(str, field.name), None)

        # Now we're left with the unprocessed fields that *must* come from
        # keywords, or default.

        for field in fields_iter:
            if kwargs:
                try:
                    val = kwargs.pop(cast(str, field.name))
                except KeyError:
                    # This is done with an exception rather than the
                    # default argument on pop because we don't want
                    # get_default() to be evaluated, and then not used.
                    # Refs #12057.
                    val = field.get_default()
            else:
                val = field.get_default()
            _setattr(self, cast(str, field.name), val)

        if kwargs and '_dn' in kwargs:
            _setattr(self, '_dn', kwargs['_dn'])
            kwargs.pop('_dn')

        if kwargs:
            for kwarg in kwargs:
                raise TypeError("'%s' is an invalid keyword argument for this function" % kwarg)
        super().__init__()
        post_init.send(sender=cls, instance=self)

    @classmethod
    def from_db(
        cls,
        attributes: List[str],
        objects: Union[LDAPData, Sequence[LDAPData]],
        many: bool = False
    ) -> Union["Model", Sequence["Model"]]:
        """
        ``objects`` is a list of raw LDAP data objects
        ``attributes`` is a
        we need to convert from the raw ldap value to the value our field stores internally

        """
        if not isinstance(objects, list):
            objects = [cast(LDAPData, objects)]
        if not many and len(objects) > 1:
            raise RuntimeError('Called {}.from_db() with many=False but len(objects) > 1'.format(
                cast(Options, cls._meta).object_name)
            )
        _attr_lookup = cast(Options, cls._meta).attribute_to_field_name_map
        _field_lookup = cast(Options, cls._meta).fields_map
        for attr in attributes:
            if attr not in _attr_lookup:
                raise FieldDoesNotExist(
                    'No field on model {} corresponding to LDAP attribute "{}"'.format(
                        cast(Options, cls._meta).object_name, attr
                    )
                )
        rows = []
        for obj in objects:
            if not isinstance(obj[1], dict):
                continue
            # Case sensitivity does not matter in LDAP, but it does when we're looking up keys in our dict here.  Deal
            # with the case for when we have a different case on our field name than what LDAP returns
            obj_attr_lookup = {k.lower(): k for k in obj[1]}
            kwargs = {}
            kwargs['_dn'] = obj[0]
            for attr in attributes:
                name = _attr_lookup[attr]
                try:
                    value: Any = obj[1][obj_attr_lookup[attr.lower()]]
                except KeyError:
                    # if the object in LDAP doesn't have that data, the
                    # attribute won't be present in the response
                    continue
                kwargs[name] = _field_lookup[name].from_db_value(value)
            rows.append(cls(**kwargs))
        if not many:
            return rows[0]
        return rows

    @classmethod
    def _default_manager(cls) -> "LdapManager":
        return cast("LdapManager", cls.objects)

    @classmethod
    def get_password_hash(cls, password: str) -> bytes:
        salt = os.urandom(8)
        h = hashlib.sha1(password.encode('utf-8'))
        h.update(salt)
        pwhash = "{SSHA}".encode('utf-8') + encode(h.digest() + salt)
        return pwhash

    def to_db(self) -> LDAPData:
        """
        ``to_db`` produces 2-tuple similar to what we would get from
        python-ldap's .search_s().::

            ( __DN__, {'attr1': ['value'], 'attr2': ['value2'], ...} )

        This data structure differs from python-ldap in that we don't prune
        attributes that have no value attached to them.  Those attributes will
        have value `[]`.

        We do this so that when ldaporm.managers.Modlist.modify() gets
        called, it can determine easily which attributes need to be deleted from
        the object in LDAP.
        """
        attrs = {}
        for field in cast(Options, self._meta).fields:
            attrs.update(field.to_db_value(field.value_from_object(self)))
        return (cast(str, self.dn), attrs)

    def __repr__(self) -> str:
        return '<%s: %s>' % (self.__class__.__name__, self)

    def __str__(self) -> str:
        return '%s object (%s)' % (self.__class__.__name__, self.dn)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Model):
            return False
        if cast(Options, self._meta).concrete_model != cast(Options, other._meta).concrete_model:
            return False
        my_pk = self.pk
        if my_pk is None:
            return self is other
        return my_pk == other.pk

    def __hash__(self) -> int:
        return hash(self.dn)

    def _get_pk_val(self, meta: Options = None) -> Any:
        _meta: Options = meta or cast(Options, self._meta)
        field = cast(Field, _meta.pk)
        return getattr(self, cast(str, field.name))

    def _set_pk_val(self, value: Any) -> None:
        _meta = cast(Options, self._meta)
        field = cast(Field, _meta.pk)
        return setattr(self, cast(str, field.name), value)

    pk = property(_get_pk_val, _set_pk_val)

    def _get_FIELD_display(self, field: Field) -> str:
        value = getattr(self, cast(str, field.name))
        return force_text(dict(field.flatchoices).get(value, value), strings_only=True)

    @property
    def dn(self) -> Optional[str]:
        if self._dn:
            return self._dn
        _meta = cast(Options, self._meta)
        manager = cast(LdapManager, _meta.base_manager)
        return manager.dn(self)

    def save(self, commit: bool = True) -> None:
        _meta = cast(Options, self._meta)
        manager = cast(LdapManager, _meta.base_manager)
        try:
            manager.get_by_dn(cast(str, self.dn))
        except self.DoesNotExist:
            manager.add(self)
        else:
            manager.modify(self)

    def delete(self) -> None:
        _meta = cast(Options, self._meta)
        manager = cast(LdapManager, _meta.base_manager)
        manager.delete_obj(self)

    def clean(self) -> None:
        """
        Hook for doing any extra model-wide validation after clean() has been
        called on every field by self.clean_fields. Any ValidationError raised
        by this method will not be associated with a particular field; it will
        have a special-case association with the field defined by NON_FIELD_ERRORS.
        """
        pass

    def full_clean(self, exclude: List[str] = None, validate_unique: bool = True) -> None:
        """
        validate_unique is here to fool ModelForm into thinking we're a Django ORM Model
        """
        errors: Dict[str, Any] = {}
        if exclude is None:
            exclude = []
        else:
            exclude = list(exclude)

        try:
            self.clean_fields(exclude=exclude)
        except ValidationError as e:
            errors = e.update_error_dict(errors)

        # Form.clean() is run even if other validation fails, so do the
        # same with Model.clean() for consistency.
        try:
            self.clean()
        except ValidationError as e:
            errors = e.update_error_dict(errors)

        if errors:
            raise ValidationError(errors)

    def clean_fields(self, exclude: List[str] = None) -> None:
        _meta = cast(Options, self._meta)
        if exclude is None:
            exclude = []

        errors: Dict[str, Any] = {}
        for f in _meta.fields:
            if f.name in exclude:
                continue
            raw_value = getattr(self, cast(str, f.name))
            if f.blank and raw_value == f.empty_values:
                continue
            try:
                setattr(self, cast(str, f.name), f.clean(raw_value, self))
            except ValidationError as e:
                errors[cast(str, f.name)] = e.error_list

        if errors:
            raise ValidationError(errors)

    def validate_unique(self, exclude: List[str] = None) -> None:
        pass



from bisect import bisect
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Type, cast

from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.utils.functional import cached_property
from django.utils.text import camel_case_to_spaces, format_lazy
from django.utils.translation import override

from .managers import LdapManager
from .fields import CharListField

if TYPE_CHECKING:
    from .models import Model  # type: ignore  # noqa:F401
    from .fields import Field

DEFAULT_NAMES = (
    'ldap_server', 'ldap_options', 'manager_class', 'basedn', 'objectclass',
    'extra_objectclasses', 'verbose_name', 'verbose_name_plural', 'ordering',
    'permissions', 'default_permissions', 'password_attribute', 'userid_attribute'
)


class Options:

    def __init__(self, meta) -> None:
        # LDAP related
        self.ldap_server: str = 'default'
        self.ldap_options: List[str] = []
        self.manager_class: Type["LdapManager"] = LdapManager
        self.basedn: Optional[str] = None
        self.objectclass: Optional[str] = None
        self.extra_objectclasses: List[str] = []
        self.userid_attribute: str = 'uid'
        self.password_attribute: Optional[str] = None

        # other
        self.verbose_name: Optional[str] = None
        self.verbose_name_plural: Optional[str] = None
        self.ordering: List[str] = []
        self.default_permissions: Tuple[str, ...] = ('add', 'change', 'delete', 'view')
        self.permissions: List[str] = []

        # these are set up by the LdapModelBase metaclass
        self.model_name: Optional[str] = None
        self.object_name: Optional[str] = None
        self.meta = meta
        self.pk: Optional["Field"] = None
        self.concrete_model: Optional["Model"] = None
        self.base_manager: Optional["LdapManager"] = None
        self.local_fields: List["Field"] = []

        # self.get_latest_by = None

        # these need to be here to fool ModelForm
        self.local_many_to_many: List["Field"] = []
        self.private_fields: List["Field"] = []
        self.many_to_many: List["Field"] = []

    @property
    def label(self) -> str:
        return cast(str, self.object_name)

    @property
    def label_lower(self) -> str:
        return cast(str, self.model_name)

    @property
    def verbose_name_raw(self) -> str:
        """Return the untranslated verbose name."""
        with override(None):
            return str(self.verbose_name)

    def contribute_to_class(self, cls: Type["Model"], name: str) -> None:
        cls._meta = self
        self.model = cls
        # First, construct the default values for these options.
        self.object_name = cls.__name__
        self.model_name = self.object_name.lower()
        self.verbose_name = camel_case_to_spaces(self.object_name)

        # Next, apply any overridden values from 'class Meta'.
        if self.meta:
            meta_attrs = self.meta.__dict__.copy()
            for attr in self.meta.__dict__:
                # NOTE: We can't modify a dictionary's contents while looping
                # over it, so we loop over the *original* dictionary instead.
                if attr.startswith('_'):
                    del meta_attrs[attr]
            for attr_name in DEFAULT_NAMES:
                if attr_name in meta_attrs:
                    setattr(self, attr_name, meta_attrs.pop(attr_name))
                elif hasattr(self.meta, attr_name):
                    setattr(self, attr_name, getattr(self.meta, attr_name))

            # verbose_name_plural is a special case because it uses a 's'
            # by default.
            if self.verbose_name_plural is None:
                self.verbose_name_plural = format_lazy('{}s', self.verbose_name)

            # Any leftover attributes must be invalid.
            if meta_attrs != {}:
                raise TypeError("'class Meta' got invalid attribute(s): %s" % ','.join(meta_attrs))
        else:
            self.verbose_name_plural = format_lazy('{}s', self.verbose_name)
        del self.meta

    def _prepare(self, model: Type["Model"]) -> None:
        if self.pk is None:
            raise ImproperlyConfigured("'{}' model doesn't have a primary key".format(self.object_name))
        # Always make sure we have objectclass in our model, so we can filter by it
        # don't call self.attributes here, because that gets cached
        for f in self._get_fields():
            if f.ldap_attribute == 'objectclass':
                raise ImproperlyConfigured(
                    "The objectclass field is defined automatically; don't manaully define it "
                    "on the '{}' model".format(self.object_name)
                )
        objectclass = CharListField('objectclass', editable=False, max_length=255)
        model.add_to_class('objectclass', objectclass)

    def add_field(self, field: "Field") -> None:
        self.local_fields.insert(bisect(self.local_fields, field), field)
        self.setup_pk(field)

    def setup_pk(self, field: "Field") -> None:
        if not self.pk and field.primary_key:
            self.pk = field

    def __repr__(self) -> str:
        return '<Options for %s>' % self.object_name

    @cached_property
    def fields(self) -> List["Field"]:
        return self._get_fields()

    @property
    def concrete_fields(self) -> List["Field"]:
        # this is here to fool ModelForm
        return self.fields

    def get_fields(self, include_parents: bool = True, include_hidden: bool = False) -> List["Field"]:
        return self._get_fields()

    def _get_fields(self) -> List["Field"]:
        return self.local_fields

    @cached_property
    def fields_map(self) -> Dict[str, "Field"]:
        res = {}
        fields = self._get_fields()
        for field in fields:
            res[cast(str, field.name)] = field
        return res

    @cached_property
    def attributes_map(self) -> Dict[str, str]:
        res = {}
        fields = self._get_fields()
        for field in fields:
            res[cast(str, field.name)] = field.ldap_attribute
        return res

    @cached_property
    def attribute_to_field_name_map(self) -> Dict[str, str]:
        return {f.ldap_attribute: cast(str, f.name) for f in self._get_fields()}

    @cached_property
    def attributes(self) -> List[str]:
        return [f.ldap_attribute for f in self._get_fields()]

    def get_field(self, field_name: str) -> "Field":
        """
        Return a field instance given the name of a forward or reverse field.
        """
        try:
            # Retrieve field instance by name from cached or just-computed
            # field map.
            return self.fields_map[field_name]
        except KeyError:
            raise FieldDoesNotExist("%s has no field named '%s'" % (self.object_name, field_name))



from typing import Dict, List, Tuple, Union

DeleteModListEntry = Tuple[int, str, None]
ModifyModListEntry = Tuple[int, str, str]
AddModlistEntry = Tuple[str, str]
ModifyDeleteModList = List[Union[DeleteModListEntry, ModifyModListEntry, AddModlistEntry]]
AddModlist = List[Tuple[str, str]]
LDAPData = Tuple[str, Dict[str, List[bytes]]]



from typing import Any, Dict, List, Optional, Type, Union, cast
from django import forms
from django.forms.widgets import Textarea, Widget
from django.forms.renderers import BaseRenderer
from django.core.exceptions import ValidationError
from django.core import validators


class CharListWidget(Textarea):

    is_hidden: bool = False

    def render(
        self,
        name: str,
        value: Optional[Union[str, List[str]]],
        attrs: Dict[str, Any] = None,
        renderer: BaseRenderer = None
    ):
        if value in self.attrs['empty_values']:
            value = ""
        elif isinstance(value, list):
            value = "\n".join(value)
        return super().render(name, value, attrs)


class CharListField(forms.CharField):

    widget: Type[Widget] = CharListWidget
    empty_values: List[Any] = list(validators.EMPTY_VALUES) + ['[]']

    def to_python(self, value: Optional[str]) -> List[str]:  # type: ignore
        if value in self.empty_values:
            return []
        res = []
        for v in cast(str, value).splitlines():
            if self.strip:
                v = v.strip()
            if v == "":
                continue
            res.append(v)
        return res

    def run_validators(self, value):
        """
        Validate each line independently.
        """
        # Use the parent's handling of required fields, etc.
        errors = []
        for v in value:
            try:
                super().run_validators(v)
            except ValidationError as e:
                errors.extend(e.error_list)
        if errors:
            raise ValidationError(errors)

    def widget_attrs(self, widget):
        attrs = super().widget_attrs(widget)
        attrs['empty_values'] = self.empty_values
        return attrs



from base64 import b64encode as encode
from collections import namedtuple
from distutils.version import StrictVersion
from functools import wraps
import hashlib
import logging
import os
import re
import threading
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type, Sequence, cast

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

import ldap
from ldap import modlist
from ldap.controls import SimplePagedResultsControl
from ldap_filter import Filter


from .typing import ModifyDeleteModList, AddModlist, LDAPData

if TYPE_CHECKING:
    from ldap_filter.filter import GroupAnd
    from .models import Model
    from .options import Options  # noqa:F401

LDAP24API = StrictVersion(ldap.__version__) >= StrictVersion('2.4')
logger = logging.getLogger('django-ldaporm')


# -----------------------
# Decorators
# -----------------------

def log_prefix(prefix: str) -> Callable:

    def real_decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(self, level, msg, args, **kwargs) -> Callable:
            msg = prefix + " " + msg
            return func(level, msg, args, **kwargs)
        return wrapper
    return real_decorator


def atomic(key: str = 'read') -> Callable:
    """
    Use this decorator to wrap methods that actually need to talk to an LDAP
    server.

    ``key`` is either "read" or "write".

    If 'key' is "read", do this operation on the LDAP server we've designated
    as our read-only server.

    If 'key' is "write", do this operation on the LDAP server we've designated
    as our read-write server.
    """
    def real_decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(self, *args, **kwargs) -> Callable:
            # add the LDAP server url to our logging context
            if self.has_connection():
                # Ensure we're not currently in a wrapped function
                return func(self, *args, **kwargs)
            old_log_method = self.logger._log
            self.logger._log = log_prefix(f"ldap_url={self.config[key]['url']}")(old_log_method)
            self.connect(key)
            try:
                retval = func(self, *args, **kwargs)
            finally:
                # We do this in a finally: branch so that the ldap
                # connection and logger gets cleaned up no matter what
                # happens in func()
                self.disconnect()
                self.logger._log = old_log_method
            return retval
        return wrapper
    return real_decorator


def substitute_pk(func: Callable) -> Callable:
    """
    Certain LdapManager() methods allow you to use the kwarg "pk".  Replace
    that with self.pk before passing into the method.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs) -> Callable:
        kw = {}
        for key, value in kwargs.items():
            if key == 'pk':
                key = self.pk
            kw[key] = value
        return func(self, *args, **kw)
    return wrapper


def needs_pk(func: Callable) -> Callable:
    """
    When we retrieve data from LDAP, in most cases we want to ensure we include
    the primary key for the object in our returned attributes so that we can
    later do .save() and .delete() on it.

    This decorator adds self.manager.pk to self._attributes before
    executing the LDAP search.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs) -> Callable:
        pk_attr = self.get_attribute(self.manager.pk)
        if pk_attr not in self._attributes:
            self._attributes.append(pk_attr)
        return func(self, *args, **kwargs)
    return wrapper


# -----------------------
# Helper Classes
# -----------------------


class Modlist:

    def __init__(self, manager: "LdapManager") -> None:
        self.manager = manager

    def _get_modlist(self, data: Dict[str, Any], modtype: int = ldap.MOD_REPLACE) -> ModifyDeleteModList:
        _modlist: ModifyDeleteModList = []
        for key in data:
            if modtype == ldap.MOD_DELETE:
                _modlist.append((ldap.MOD_DELETE, key, None))
            elif modtype == ldap.MOD_ADD:
                _modlist.append((key, data[key]))
            else:
                _modlist.append((cast(int, modtype), key, data[key]))
        return _modlist

    def add(self, obj: "Model") -> AddModlist:
        """
        Convert an LDAP DAO object to a modlist suitable for passing to
        `add_s` and return it.
        """
        data = obj.to_db()
        if hasattr(obj, 'objectclass'):
            data[1]['objectclass'] = obj.objectclass  # type: ignore
        else:
            raise ImproperlyConfigured("Tried to add an object with no objectclasses defined.")
        # We have to do these two casts because LdapManager.model and
        # Model._meta start out as None.  By the time we get here the metaclass
        # has filled in those two so we know they're not None, but mypy has no
        # way of knowing that
        model = cast("Model", self.manager.model)
        _meta = cast("Options", model._meta)
        _attribute_lookup = _meta.attribute_to_field_name_map
        _fields_map = _meta.fields_map
        # purge the empty fields
        new = {}
        for key, value in data[1].items():
            field = _fields_map[_attribute_lookup[key]]
            if not field.editable and not key == 'objectclass':
                continue
            if value != []:
                new[key] = value
        return modlist.addModlist(new)

    def update(self, new: "Model", old: "Model", force: bool = False) -> ModifyDeleteModList:
        """
        We do our own implementation of `ldap.modify.modifyModlist` here because
        python-ldap explicitly says:

            Replacing attribute values is always done with a
            ldap.MOD_DELETE/ldap.MOD_ADD pair instead of ldap.MOD_REPLACE to
            work-around potential issues with attributes for which no EQUALITY
            matching rule are defined in the server’s subschema.

            (see https://www.python-ldap.org/en/latest/reference/ldap-modlist.html)

        Our version figures out whether to ldap.MOD_DELETE or ldap.MOD_REPLACE
        based on whether the value is empty.  This is needed by the 389 family
        servers because they now enforce rules for attributes that are optional
        but can't have a null value.

        :param new: the data for the object we have
        :type new: an LDAP ORM object

        :param old: the data for the object currently in LDAP
        :type old: an LDAP ORM object

        :rtype: list
        """
        # We have to do these two casts because LdapManager.model and
        # Model._meta start out as None.  By the time we get here the metaclass
        # has filled in those two so we know they're not None, but mypy has no
        # way of knowing that
        model = cast("Model", self.manager.model)
        _meta = cast("Options", model._meta)
        _attribute_lookup = _meta.attribute_to_field_name_map
        _fields_map = _meta.fields_map
        # first build the changed attributes
        old_data = old.to_db()
        new_data = new.to_db()
        changes = {}
        for key, value in new_data[1].items():
            field = _fields_map[_attribute_lookup[key]]
            if not field.editable:
                continue

            if old_data[1][key] != value:
                changes[key] = value

        # Now build the ldap.MOD_DELETE and ldap.MOD_REPLACE modlists
        deletes: Dict[str, Any] = {}
        replacements: Dict[str, Any] = {}
        for key, value in changes.items():
            if value == [] or all(x is None for x in value):
                deletes[key] = None
            else:
                replacements[key] = value
        d_modlist = self._get_modlist(deletes, ldap.MOD_DELETE)
        r_modlist = self._get_modlist(replacements, ldap.MOD_REPLACE)
        return r_modlist + d_modlist


# ========================================
# The LDAP search filter QuerySet analog
# ========================================


class F:

    # need to be able to specify field name and lookup db_column in LDAP

    class NoFilterSpecified(Exception):
        pass

    class UnknownSuffix(Exception):
        pass

    class UnboundFilter(Exception):
        pass

    def __init__(self, manager: "LdapManager", f: "F" = None) -> None:
        self.manager = manager
        self.model = cast(Type["Model"], manager.model)
        self._meta = cast("Options", self.model._meta)
        self.fields_map = self._meta.fields_map
        self.attributes_map = self._meta.attributes_map
        self.attribute_to_field_name_map = self._meta.attribute_to_field_name_map
        self.attributes = self._meta.attributes
        self._attributes = self.attributes
        self._order_by = self._meta.ordering
        if f:
            self.chain: List["F"] = f.chain
        else:
            self.chain = []

    @property
    def _filter(self) -> "GroupAnd":
        """
        Return a list of filters ready to be converted to a filter string.

        This means throwing an "(& )" around the list of filter components
        we've accrued.
        """
        if len(self.chain) == 0:
            raise self.NoFilterSpecified('You need to at least specify one filter in order to do LDAP searches.')
        if len(self.chain) == 1:
            return self.chain[0]
        return Filter.AND(self.chain).simplify()

    def __sort(self, objects: Sequence["Model"]) -> Sequence["Model"]:
        """
        This is called by methods that return lists of results.  Sort our
        ``objects``, a list of objects of class ``self.manager.model`` based
        on ``self._order_by``.

        Example::

            self.__sort(objects, ('-pub_date', 'headline',))

        The result above will be ordered by ``pub_date`` descending, then by
        ``headline ascending``. The negative sign in front of "-pub_date" indicates
        descending order.
        """
        if not self._order_by:
            return objects
        if not any(k.startswith('-') for k in self._order_by):
            # if none of the keys are reversed, just sort directly
            return sorted(objects, key=lambda obj: tuple(getattr(obj, k) for k in self._order_by))
        # At least one key was reversed. now we have to do it the hard way,
        # with sequential sorts, starting from the last sort key and working
        # our way back to the first
        keys = list(self._order_by)
        keys.reverse()
        for k in keys:
            key = k
            reverse = False
            if key.startswith('-'):
                key = k[1:]
                reverse = True
            data = sorted(
                objects,
                key=lambda obj: getattr(obj, key),  # pylint: disable=cell-var-from-loop
                reverse=reverse
            )
        return data

    def __validate_positional_args(self, args: Sequence["F"]) -> List["F"]:
        if args:
            for arg in args:
                if not isinstance(arg, F):
                    raise ValueError(
                        "F.filter() positional arguments must all be F() objects."
                    )
        return list(args)

    def get_attribute(self, name: str) -> str:
        try:
            return self.attributes_map[name]
        except KeyError:
            raise self.model.InvalidField(
                '"{}" is not a valid field on model {}'.format(name, self.model.__name__)
            )

    def wildcard(self, name: str, value: str) -> "F":
        """
        Convert ``value`` with some "*" in it (beginning, end or both) to
        appropriate LDAP filters ("__iendswith", "__istartswith" or
        "__icontains"), apply the the filter to our current state and return the
        result.

        If there are no "*" characters at the beginning or end of ``value`` do
        an "__iexact"" filter.

        :param name: the the name of one of the fields on our model
        :type name: string

        :param value: the keyword to look for
        :type value: string

        :rtype: a core.ldap.managers.F object
        """
        value = value.strip()
        f = self
        if value:
            if value.startswith('*') and value.endswith('*'):
                qualifier = "{}__icontains".format(name)
            elif value.startswith('*'):
                qualifier = "{}__iendswith".format(name)
            elif value.endswith('*'):
                qualifier = "{}__istartswith".format(name)
            else:
                qualifier = "{}__iexact".format(name)
            d = {qualifier: re.sub(r'[*]', '', value)}
            f = self.filter(**d)
        return f

    def filter(self, *args: "F", **kwargs) -> "F":
        """
        If there are positional arguments, they must all be F() objects.  This
        allows us to preconstruct an unbound filter and then use it later.

        Of the Django filter suffixes we support the following:

            * ``__iexact``
            * ``__istartswith``
            * ``__iendswith``
            * ``__icontains``
            * ``__in``

        Note that only the case insensitive versions of these filters are supported.
        LDAP searches are case insensitive, so we make you use ``__i*`` versions to
        remind you of that.

        We do support one additional LDAP specific filter: ``__exists``.  This
        will cause a filter to be added that just ensures that the returned
        objects have the associated attribute.  Unlike in SQL, where every row
        contains all available columns, in LDAP, attributes can be absent
        entirely from the record.
        """
        steps = self.__validate_positional_args(args)
        for key, value in kwargs.items():
            if isinstance(value, str):
                value = value.strip()
            if key.endswith('__istartswith'):
                steps.append(Filter.attribute(self.get_attribute(key[:-13])).starts_with(value))
            elif key.endswith('__iendswith'):
                steps.append(Filter.attribute(self.get_attribute(key[:-11])).ends_with(value))
            elif key.endswith('__icontains'):
                steps.append(Filter.attribute(self.get_attribute(key[:-11])).contains(value))
            elif key.endswith('__in'):
                if not isinstance(value, list):
                    raise ValueError('When using the "__in" filter you must supply a list')
                in_steps = []
                for v in value:
                    in_steps.append(Filter.attribute(self.get_attribute(key[:-4])).equal_to(v))
                steps.append(Filter.OR(in_steps))
            elif key.endswith('__exists'):
                # This one doesn't exist as a Django field lookup
                steps.append(Filter.attribute(self.get_attribute(key[:-8])).present())
            elif key.endswith('__iexact') or '__' not in key:
                # no suffix means do an __exact
                if '__' in key:
                    key = key[:-8]
                if value is None:
                    # If value is None, we search for the absence of that
                    # attribute
                    steps.append(Filter.NOT(Filter.attribute(self.get_attribute(key))))
                else:
                    steps.append(Filter.attribute(self.get_attribute(key)).equal_to(value))
            else:
                raise F.UnknownSuffix('The search filter "{}" uses an unknown filter suffix')
        self.chain.append(Filter.AND(steps))
        return self

    def only(self, *names) -> "F":
        """
        Return an object with only these attributes.  Any attribute in the list which
        is not in the set of known attributes for our model will cause us to raise
        self.manager.model.InvalidField.

        For multiple calls to .only(), the last call wins.

        :param names: list of field names to retrieve from LDAP for this object
        :type names: list ot strings
        """
        self._attributes = [self.get_attribute(name) for name in names]
        return self

    @needs_pk
    def first(self) -> "Model":
        """
        The difference between .first() and .get() is that with .first() we know
        we might get more than one result, but we just want the first one.

        With .get() we expect that there should only be only one result and so if
        our search returns more, we want to know.

        .. note::

            If you'll don't have an .order_by() filter in your stack of filters,
            you'll get the first object from the search directly from LDAP.
            LDAP orders objects internally by insertion order, so you'll just
            get whatever the oldest object in LDAP is that matches your filters.
            But, this allows us to ask LDAP for just one object, which is fast.

            If you do have an .order_by() filter, you'll get the first object after
            sorting.  But LDAP servers won't do any sorting for you, so in this
            case we retrieve all objects matching our search filters and just return
            the first one, which may be expensive.
        """
        sizelimit = 0
        if not self._order_by:
            # We can just take the default LDAP ordering, so just take the first
            # result in whatever order the LDAP server keeps it in.
            sizelimit = 1
        objects = self.manager.search(str(self), self._attributes, sizelimit=sizelimit)
        if len(objects) == 0:
            raise self.model.DoesNotExist(
                'A {} object matching query does not exist.'.format(self.model.__name__))
        objects = self.model.from_db(self._attributes, objects)
        if not self._order_by:
            return objects[0]
        return self.__sort(objects)[0]

    @needs_pk
    def get(self) -> "Model":
        objects = self.manager.search(str(self), self._attributes)
        if len(objects) == 0:
            raise self.model.DoesNotExist(
                'A {} object matching query does not exist.'.format(self.model.__name__))
        if len(objects) > 1:
            raise self.model.MultipleObjectsReturned(
                'More than one {} object matched query.'.format(self.model.__name__))
        return cast("Model", self.model.from_db(self._attributes, objects))

    @needs_pk
    def update(self, **kwargs) -> None:
        obj = self.get()
        # FIXME: obj.dump() does not exist;
        new = self.model.from_db(self._attributes, obj.dump())
        for key, value in kwargs.items():
            if key in self.attributes:
                setattr(new, key, value)
        self.manager.modify(new, old=obj)

    def exists(self) -> bool:
        """
        Return ``True`` if the LDAP search with the filter we've built returns
        any results, ``False`` if not.

        :rtype: boolean
        """
        objects = self.manager.search(str(self), self._attributes)
        return len(objects) > 0

    @needs_pk
    def all(self) -> Sequence["Model"]:
        objects = self.model.from_db(
            self._attributes,
            self.manager.search(str(self), self._attributes),
            many=True
        )
        return self.__sort(cast(Sequence["Model"], objects))

    def delete(self) -> None:
        """
        Delete an object that matches our filters.

        .. note::

            This works differently than Django's QuerySet .delete() function in
            that it will allow you to delete one and only one object from LDAP.
            The filters you pass should uniquely identify a single LDAP object.

            We do this to protect LDAP itself from our bugs.
        """
        obj = self.get()
        self.manager.connection.delete_s(obj.dn)

    def order_by(self, *args: str) -> "F":
        """
        When we return results, order them by the positional arguments
        Example::

            Entry.objects.order_by('-pub_date', 'headline').all()

        The result above will be ordered by ``pub_date`` descending, then by
        ``headline ascending``. The negative sign in front of "-pub_date" indicates
        descending order.
        """
        for key in args:
            _key = key[1:] if key.startswith('-') else key
            if _key not in self.attributes_map:
                raise self.model.InvalidField(
                    '"{}" is not a valid field on model {}'.format(_key, self.model.__name__)
                )
        self._order_by = list(args)
        return self

    def values(self, *attrs: str) -> List[Dict[str, Any]]:
        """
        Returns a a list of dictionaries, rather than of model instances.  Each
        of those dictionaries represents an object, with the keys corresponding
        to the attribute names of model objects.

            >>> Entry.objects.values('uid', 'sn')
            [{'uid': 'Barney', 'sn': 'Rubble'}, {'uid': 'Fred', 'sn': 'Flintstone'}, ...]

        The ``values()`` method takes optional positional arguments, ``*attrs``,
        which specify attribute names should be included in the dictionaries.
        If you specify the attributes, each dictionary will contain only the
        attribute keys/values for the fields you specify. If you don’t specify
        the attribute, each dictionary will contain a key and value for every
        attribute defined on the model.

        .. note::

            I'm trying to model how the Django ORM works here, so ``.values()``
            is not compatible with ``.only()``.  If previously in your filter
            chain you have an ``.only()``, you'll get  a ``NotImplementedError``
            exception.

        """
        if self._attributes != self.attributes:
            raise NotImplementedError("Don't use .only() with .values()")
        _attrs = []
        if not attrs:
            _attrs = self.attributes
        objects = self.model.from_db(_attrs, self.manager.search(str(self), _attrs), many=True)
        objects = self.__sort(cast(Sequence["Model"], objects))
        data = []
        for obj in objects:
            data.append({self.attribute_to_field_name_map[attr]: getattr(obj, attr) for attr in _attrs})
        return data

    def values_list(self, *attrs: str, **kwargs) -> List[Tuple[Any, ...]]:
        """
        This is similar to values() except that instead of returning
        dictionaries, it returns a list of tuples.  Each tuple contains
        the value from the respective field or expression passed into the
        ``values_list()`` call — so the first item is the first field, etc. For
        example::

            >>> Entry.objects.values_list('uid', 'sn')
            [('Barney', 'Rubble'), ('Fred', 'Flintstone'), ...]

        If you only pass in a single field, you can also pass in the ``flat``
        parameter. If ``True``, this will mean the returned results are a list
        of single values, rather than a list of one-tuples.

            >>> Entry.objects.values_list('uid', flat=True)
            ['barney', 'fred', ...]

        It is an error to pass in flat when there is more than one field.

        You can pass named=True to get results as a ``namedtuple()``.

            >>> Entry.objects.values_list('uid', named=True)
            [Row(uid='barney'), Row(uid='fred'), ...]

        .. note::

            I'm trying to model how the Django ORM works here, so ``.values()``
            is not compatible with ``.only()``.  If previously in your filter
            chain you have an ``.only()``, you'll get  a ``NotImplementedError``
            exception.
        """
        if self._attributes != self.attributes:
            raise NotImplementedError("Don't use .only() with .values_list()")

        _attrs: List[str] = []
        if not attrs:
            _attrs = self.attributes
            attrs = tuple(self.attribute_to_field_name_map[attr] for attr in _attrs)
        else:
            _attrs = [self.get_attribute(attr) for attr in attrs]
        objects = self.model.from_db(_attrs, self.manager.search(str(self), _attrs), many=True)
        objects = self.__sort(cast(Sequence["Model"], objects))
        if 'flat' in kwargs and kwargs['flat']:
            if len(attrs) > 1:
                raise ValueError("Cannot use flat=True when asking for more than one field")
            return [getattr(obj, attrs[0]) for obj in objects]
        if 'named' in kwargs and kwargs['named']:
            rows: List[Any] = []
            for obj in objects:
                Row = namedtuple('Row', attrs)  # type: ignore
                # the keys here should be field names, not attribute names
                rows.append(Row(**{attr: getattr(obj, attr) for attr in attrs}))
            return rows
        data: List[Tuple[str, ...]] = []
        for obj in objects:
            data.append(tuple(getattr(obj, attr) for attr in attrs))
        return data

    def __or__(self, other: "F") -> "F":
        self.chain = Filter.OR([self._filter, other._filter])
        return F(self.manager, f=self)

    def __and__(self, other: "F") -> "F":
        self.chain = Filter.AND([self._filter, other._filter])
        return F(self.manager, f=self)

    def __str__(self) -> str:
        return self._filter.to_string()


# -----------------------
# LdapManager
# -----------------------


class LdapManager:

    def __init__(self) -> None:
        """
        This class does all of the direct interactions with LDAP and should be
        the only one that calls the ``ldap`` library functions.
        """
        self.logger = logger
        self.pagesize: int = 100

        # These get set during contribute_to_class()
        # self.config is the part of settings.LDAP_SERVERS that we need for our Model
        self.config: Optional[Dict[str, Any]] = None
        self.model: Optional[Type["Model"]] = None
        self.pk: Optional[str] = None
        self.ldap_options: List[str] = []
        self.objectclass: Optional[str] = None
        self.extra_objectclasses: List[str] = []

        # keys in this dictionary get manipulated by .connect() and
        # .disconnect()
        self._ldap_objects: Dict[threading.Thread, ldap.ldapobject.LDAPObject] = {}

    def _get_pctrls(self, serverctrls):
        """
        Lookup an LDAP paged control object from the returned controls.
        """
        # Look through the returned controls and find the page controls.
        # This will also have our returned cookie which we need to make
        # the next search request.
        return [c for c in serverctrls if c.controlType == SimplePagedResultsControl.controlType]

    def _paged_search(
        self,
        basedn: str,
        searchfilter: str,
        attrlist: List[str] = None,
        pagesize: int = 100,
        sizelimit: int = 0,
        scope: int = ldap.SCOPE_SUBTREE
    ) -> List[LDAPData]:
        """
        Performs a pages search against the LDAP server. Code lifted from:
        https://gist.github.com/mattfahrner/c228ead9c516fc322d3a
        """
        # Initialize the LDAP controls for paging. Note that we pass ''
        # for the cookie because on first iteration, it starts out empty.
        controls = SimplePagedResultsControl(True, size=pagesize, cookie='')

        # Do searches until we run out of pages to get from the LDAP server.
        results: List[LDAPData] = []
        while True:
            # Send search request.
            msgid = self.connection.search_ext(
                basedn,
                scope,
                searchfilter,
                attrlist,
                serverctrls=[controls],
                sizelimit=sizelimit
            )
            rtype, rdata, rmsgid, serverctrls = self.connection.result3(msgid)  # pylint: disable=unused-variable
            # Each "rdata" is a tuple of the form (dn, attrs), where dn is
            # a string containing the DN (distinguished name) of the entry,
            # and attrs is a dictionary containing the attributes associated
            # with the entry. The keys of attrs are strings, and the associated
            # values are lists of strings.
            for dn, attrs in rdata:

                # AD returns an rdata at the end that is a reference that we want to ignore
                if isinstance(attrs, dict):
                    results.append((dn, attrs))

            # Get cookie for the next request.
            paged_controls = self._get_pctrls(serverctrls)
            if not paged_controls:
                self.logger.warning(
                    f'paged_search.rfc2696_control_ignored searchfilter={searchfilter} '
                    'pagesize={pagesize]} sizelimit={sizelimit}',
                )
                break

            # Push cookie back into the main controls.
            controls.cookie = paged_controls[0].cookie

            # If there is no cookie, we're done!
            if not paged_controls[0].cookie:
                break
        return results

    def contribute_to_class(self, cls, accessor_name):
        self.pk = cls._meta.pk.name
        self.basedn = cls._meta.basedn
        self.objectclass = cls._meta.objectclass
        self.extra_objectclasses = cls._meta.extra_objectclasses
        self.ldap_options = cls._meta.ldap_options
        try:
            self.config = settings.LDAP_SERVERS[cls._meta.ldap_server]
        except AttributeError:
            raise ImproperlyConfigured("settings.LDAP_SERVERS does not exist!")
        except KeyError:
            raise ImproperlyConfigured(
                "{}: settings.LDAP_SERVERS has no key '{}'".format(
                    cls.__name__,
                    cls._meta.ldap_server,
                )
            )

        if not self.basedn:
            try:
                self.basedn = self.config['basedn']
            except KeyError:
                raise ImproperlyConfigured(
                    "{}: no ``Meta.basedn`` and settings.LDAP_SERVERS['{}'] has no 'basedn' key".format(
                        cls.__name__,
                        cls._meta.ldap_server
                    )
                )
        self.model = cls
        cls._meta.base_manager = self
        setattr(cls, accessor_name, self)

    def __get_dn_key(self, meta: "Options"):
        _attribute_lookup = meta.attribute_to_field_name_map
        dn_key = self.pk
        for k, v in _attribute_lookup.items():
            if v == self.pk:
                dn_key = k
                break
        return dn_key

    def dn(self, obj: "Model") -> Optional[str]:
        if not obj._dn:
            dn_key = self.__get_dn_key(cast("Options", obj._meta))
            pk_value = getattr(obj, cast(str, self.pk))
            if pk_value:
                obj._dn = "{}={},{}".format(dn_key, getattr(obj, cast(str, self.pk)), self.basedn)
            else:
                # If our pk_value is None or '', we're in the middle of creating a new record and haven't set
                # it yet.
                obj._dn = None
        return obj._dn

    def get_dn(self, pk: str) -> str:
        """
        Given a value for an object primary key, return what the dn for that
        object would look like.

        Args:
            pk: the value for the primary key for our model

        Returns:
            A fully qualified dn.
        """
        meta = cast("Options", cast("Model", self.model)._meta)
        dn_key = self.__get_dn_key(meta)
        return "{}={},{}".format(dn_key, pk, self.basedn)

    def disconnect(self) -> None:
        self.connection.unbind_s()
        self.remove_connection()

    def has_connection(self) -> bool:
        return threading.current_thread() in self._ldap_objects

    def set_connection(self, obj: ldap.ldapobject.LDAPObject) -> None:
        self._ldap_objects[threading.current_thread()] = obj

    def remove_connection(self) -> None:
        del self._ldap_objects[threading.current_thread()]

    def _connect(
        self,
        key: str,
        dn: str = None,
        password: str = None
    ) -> ldap.ldapobject.LDAPObject:
        config = cast(Dict[str, Any], self.config)[key]
        if not dn:
            dn = config['user']
            password = config['password']
        ldap_object: ldap.ldapobject.LDAPObject = ldap.initialize(config['url'])
        ldap_object.set_option(ldap.OPT_REFERRALS, 0)
        ldap_object.set_option(ldap.OPT_NETWORK_TIMEOUT, 15.0)
        ldap_object.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
        ldap_object.set_option(ldap.OPT_X_TLS_NEWCTX, 0)
        ldap_object.start_tls_s()
        ldap_object.simple_bind_s(dn, password)
        return ldap_object

    def connect(
        self,
        key: str,
        dn: str = None,
        password: str = None
    ) -> None:
        """
        This is used internally to set our per-thread connection object.  It is primarily used by
        the ``@atomic`` decorator.

        Args:
            key: A key into our ``self.settings`` configuration object.  This
                holds has hostname, bind dn, password and basedn for our read-only
                and read-write servers.

        Keyword Args:
            dn: If provided, use this as our bind dn instead of the dn in our ``self.settings``
                configuration object.
            password: If provided, use this as our password instead of the
                password in our ``self.settings`` configuration object.

        """
        self._ldap_objects[threading.current_thread()] = self._connect(key, dn=dn, password=password)

    def new_connection(
        self,
        key: str = 'read',
        dn: str = None,
        password: str = None
    ) -> ldap.ldapobject.LDAPObject:
        """
        This is used internally to set our per-thread connection object.  It is primarily used by
        the ``@atomic`` decorator.

        Keyword Args:
            key: A key into our ``self.settings`` configuration object.  This
                holds has hostname, bind dn, password and basedn for our read-only
                and read-write servers.
            dn: If provided, use this as our bind dn instead of the dn in our ``self.settings``
                configuration object.
            password: If provided, use this as our password instead of the
                password in our ``self.settings`` configuration object.

        Returns:
            A properly connected python-ldap ``LDAPObject``.
        """
        return self._connect(key, dn=dn, password=password)

    @property
    def connection(self) -> ldap.ldapobject.LDAPObject:
        return self._ldap_objects[threading.current_thread()]

    def _get_ssha_hash(self, password: str) -> bytes:
        salt = os.urandom(8)
        h = hashlib.sha1(password.encode('utf-8'))
        h.update(salt)
        pwhash = "{SSHA}".encode('utf-8') + encode(h.digest() + salt)
        return pwhash

    @atomic(key='read')
    def search(
        self,
        searchfilter: str,
        attributes: List[str],
        sizelimit: int = 0,
        basedn: str = None,
        scope: int = ldap.SCOPE_SUBTREE
    ) -> List[LDAPData]:
        if basedn is None:
            basedn = self.basedn
        if 'paged_search' in self.ldap_options:
            return self._paged_search(
                basedn,
                searchfilter,
                attrlist=attributes,
                sizelimit=sizelimit,
                scope=scope
            )
        # We have to filter out and references that AD puts in
        data = self.connection.search_s(
            basedn,
            scope,
            filterstr=searchfilter,
            attrlist=attributes
        )
        objects = []
        for obj in data:
            if isinstance(obj[1], dict):
                objects.append(obj)
        return objects

    @atomic(key='write')
    def add(self, obj: "Model") -> None:
        # This is a bit weird here because the objectclass CharListField gets
        # secretly added during class construction on Model
        obj.objectclass = []  # type: ignore
        for objectclass in self.extra_objectclasses:
            obj.objectclass.append(objectclass.encode())  # type: ignore
        obj.objectclass.append(self.objectclass.encode())  # type: ignore
        _modlist = Modlist(self).add(obj)
        self.connection.add_s(self.dn(obj), _modlist)

    @atomic(key='write')
    @substitute_pk
    def delete(self, *args, **kwargs) -> None:
        """
        Delete an object that matches our filters.

        .. note::

            This works differently than Django's QuerySet .delete() function in
            that it will allow you to delete one and only one object from LDAP.
            The filters you pass should uniquely identify a single LDAP object.

            We do this to protect LDAP itself from our bugs.
        """
        obj = self.filter(*args, **kwargs).only(self.pk).get()
        self.connection.delete_s(obj.dn)

    @atomic(key='write')
    def delete_obj(self, obj: "Model") -> None:
        """
        Delete a specified object.
        """

        self.connection.delete_s(obj.dn)

    @atomic(key='write')
    def rename(self, old_dn: str, new_dn: str) -> None:
        """
        Update an object's dn, keeping it within the same basedn.
        """
        newrdn = new_dn.split(',')[0]
        old_basedn = ','.join(old_dn.split(',')[1:])
        new_basedn = ','.join(new_dn.split(',')[1:])
        newsuperior = None
        if old_basedn != new_basedn:
            newsuperior = new_basedn
        self.connection.rename_s(old_dn, newrdn, newsuperior)

    @atomic(key='write')
    def modify(self, obj: "Model", old: "Model" = None) -> None:
        # First check to see whether we updated our primary key.  If so, we need to rename
        # the object in LDAP, and its obj._dn.  The old obj._dn should reference the old PK.
        # We'll .lower() them to deal with case for the pk in the dn
        old_pk_value = cast(str, obj.dn).split(',')[0].split('=')[1].lower()
        new_pk_value = getattr(obj, cast(str, self.pk)).lower()
        if new_pk_value != old_pk_value:
            # We need to do a modrdn_s if we change pk, to cause the dn to be updated also
            self.connection.modrdn_s(obj.dn, f'{self.pk}={new_pk_value}')
            # And update our object's _dn to the new one
            basedn = ','.join(cast(str, obj.dn).split(',')[1:])
            new_dn = f'{self.pk}={new_pk_value},{basedn}'
            obj._dn = new_dn
            # force reload old, if it was passed in so that we get the new pk value and dn
            old = None

        # Now update the non-PK attributes
        if not old:
            pk_val = cast(str, self.pk)
            old = self.get(**{pk_val: getattr(obj, pk_val)})
        _modlist = Modlist(self).update(obj, cast("Model", old))
        if _modlist:
            # Only issue the modify_s if we actually have changes
            self.connection.modify_s(obj.dn, _modlist)
        else:
            self.logger.debug('ldaporm.manager.modify.no-changes dn=%s', obj.dn)

    def only(self, *names: str) -> "F":
        return F(self).only(*names)

    def __filter(self) -> "F":
        f = F(self)
        if self.objectclass:
            f = f.filter(objectclass=self.objectclass)
        return f

    @substitute_pk
    def wildcard(self, name: str, value: str) -> "F":
        return self.__filter().wildcard(name, value)

    @substitute_pk
    def filter(self, *args, **kwargs) -> "F":
        return self.__filter().filter(*args, **kwargs)

    def get_by_dn(self, dn: str) -> "Model":
        """
        Get an object specifically by its DN.  To do this we do a search with
        the basedn set to the dn of the object, with scope ``ldap.SCOPE_BASE``
        and then get all objects that match.  This will be either the object
        we're looking for, or nothing.

        Args:
            dn: the dn to search for

        Raises:
            ValueError: `dn` is not in our Model's basedn
            Model.DoesNotExist: no object with this dn exist in our LDAP server
            Model.MultipleObjectsReturned: something really wrong has happened

        Returns:
            The ldap object corresponding to the dn
        """
        dn = dn.lower()
        model = cast(Type["Model"], self.model)
        options = cast("Options", model._meta)
        if options.basedn and not dn.endswith(options.basedn.lower()):
            raise ValueError(f"The requested dn '{dn}' is not in our model's basedn '{options.basedn}")
        try:
            objects = self.search(
                '(objectClass=*)',
                options.attributes,
                basedn=dn,
                scope=ldap.SCOPE_BASE
            )
        except ldap.NO_SUCH_OBJECT:    # pylint:disable=no-member
            objects = []
        if len(objects) == 0:
            raise model.DoesNotExist(
                'A {} object matching query does not exist.'.format(model.__name__))
        if len(objects) > 1:
            raise model.MultipleObjectsReturned(
                'More than one {} object matched query.'.format(model.__name__))
        return cast("Model", model.from_db(options.attributes, objects))

    @substitute_pk
    def get(self, *args, **kwargs) -> "Model":
        return self.__filter().filter(*args, **kwargs).get()

    def all(self) -> Sequence["Model"]:
        """
        .. note::

            Unlike the Django QuerySet version of this, this actually
            runs the query against LDAP and returns the result.  The Django
            version returns an iterator, I think.
        """
        return self.__filter().all()

    def values(self, *args: str) -> List[Dict[str, Any]]:
        return self.__filter().values(*args)

    def values_list(self, *args: str, **kwargs) -> List[Tuple[Any, ...]]:
        return self.__filter().values_list(*args, **kwargs)

    def order_by(self, *args: str) -> "F":
        return self.__filter().order_by(*args)

    def reset_password(self, username: str, new_password: str, attributes: Dict[str, Any] = None) -> bool:
        model = cast(Type["Model"], self.model)
        password_attribute = cast("Options", model._meta).password_attribute
        if not password_attribute:
            return False
        if not attributes:
            attributes = {}
        try:
            user = self.filter(**{'uid': username}).only('uid').get()
        except model.DoesNotExist:
            self.logger.warning('auth.no_such_user user=%s', username)
            return False

        pwhash = model.get_password_hash(new_password)
        attr = {password_attribute: [pwhash]}
        cast(Dict[str, Any], attributes).update(attr)

        _modlist = Modlist(self)._get_modlist(attr, ldap.MOD_REPLACE)

        self.connect('write')
        self.connection.modify_s(user.dn, _modlist)
        self.disconnect()
        service = getattr(model._meta, 'ldap_server', 'ldap')
        self.logger.info('%s.password_reset.success dn=%s', service, user.dn)
        return True

    def authenticate(self, username: str, password: str) -> bool:
        """
        Try to authenticate a username/password vs our LDAP server.

        If the user does not exist in LDAP, return False.
        If the user exists, but the bind fails, return False.
        Else, return True.

        :param username: a bare username for the person trying to authenticate e.g. 'fred'
        :type username: string

        :param password: the password to try to authenticate with
        :type password: string

        :param remote_ip: if provided, include this IP address in our log messages
        :type remote_ip: string

        :rtype: boolean
        """
        model = cast(Type["Model"], self.model)
        uid_attr = cast("Options", model._meta).userid_attribute
        try:
            user = self.filter(**{uid_attr: username}).only(uid_attr).get()
        except model.DoesNotExist:
            self.logger.warning('auth.no_such_user user=%s', username)
            return False
        try:
            self.connect('read', user.dn, password)
        except ldap.INVALID_CREDENTIALS:
            self.logger.warning('auth.invalid_credentials user=%s', username)
            return False
        self.disconnect()
        self.logger.info('auth.success user=%s', username)
        return True

    def create(self, **kwargs) -> "Model":
        """
        Create a model object based on **kwargs, then LDAP_ADD it to LDAP.
        """
        obj = cast(Type["Model"], self.model)(**kwargs)
        self.add(obj)
        return self.get(pk=getattr(obj, cast(str, self.pk)))



from base64 import b64encode as encode

import collections.abc
import datetime
from functools import partialmethod, total_ordering
import hashlib
import os
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Sequence, Tuple, Type, Optional, Union, cast
import warnings

import pytz

from django import forms
from django.conf import settings
from django.core import checks, exceptions, validators as dj_validators
from django.db.models.constants import LOOKUP_SEP
from django.db.models.fields import BLANK_CHOICE_DASH, NOT_PROVIDED, return_None  # type: ignore
from django.utils import timezone
from django.utils.dateparse import (
    parse_date,
    parse_datetime
)
from django.utils.functional import cached_property
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _

from . import forms as ldap_forms
from .validators import validate_email_forward

if TYPE_CHECKING:
    from .models import Model


Validator = Callable[[Any], None]


@total_ordering
class Field:

    """
    This is enough of a django.db.models.fields.Field implementation to allow us
    to build Django ORM-like models and fool ModelForm into working for us.
    """

    empty_strings_allowed: bool = True
    empty_values: List[Any] = list(dj_validators.EMPTY_VALUES)
    creation_counter: int = 0

    default_validators: List[Validator] = []  # Default set of validators
    default_error_messages: Dict[str, str] = {
        'invalid_choice': _('Value %(value)r is not a valid choice.'),
        'null': _('This field cannot be null.'),
        'blank': _('This field cannot be blank.'),
    }

    # These are here to fool ModelForm into thinking we're a Django ORM Field. We
    # don't actually use them.
    many_to_many: Any = None
    many_to_one: Any = None
    one_to_many: Any = None
    one_to_one: Any = None
    related_model: Any = None

    # Field flags
    hidden: bool = False

    # Generic field type description, usually overridden by subclasses
    def _description(self) -> str:
        return _('Field of type: %(field_type)s') % {
            'field_type': self.__class__.__name__
        }
    description = property(_description)

    def __init__(
        self,
        verbose_name: str = None,
        name: str = None,
        primary_key: bool = False,
        max_length: int = None,
        blank: bool = False,
        null: bool = False,
        default: Any = NOT_PROVIDED,
        editable: bool = True,
        choices: List[Any] = None,
        help_text: str = '',
        validators: Sequence[Validator] = (),
        error_messages: Dict[str, str] = None,
        db_column: str = None
    ) -> None:
        self.name = name
        self.verbose_name = verbose_name  # May be set by set_attributes_from_name
        self.primary_key = primary_key
        self.max_length = max_length
        self.blank, self.null = blank, null
        self.default = default
        self.editable = editable
        if isinstance(choices, collections.abc.Iterator):
            choices = list(choices)
        self.choices: List[Any] = choices or []
        self.help_text = help_text
        self.blank = True
        self.db_column = db_column

        self.model: Optional[Type["Model"]] = None

        self.creation_counter = Field.creation_counter
        Field.creation_counter += 1

        self._validators = list(validators)  # Store for deconstruction later

        messages: Dict[str, str] = {}
        for c in reversed(self.__class__.__mro__):
            messages.update(getattr(c, 'default_error_messages', {}))
        messages.update(error_messages or {})
        self.error_messages = messages

    def __repr__(self) -> str:
        """Display the module, class, and name of the field."""
        path = '%s.%s' % (self.__class__.__module__, self.__class__.__qualname__)
        name = getattr(self, 'name', None)
        if name is not None:
            return '<%s: %s>' % (path, name)
        return '<%s>' % path

    def __lt__(self, other: "Field") -> bool:
        """
        We need to implement this because django.forms.models.fields_from_model
        tries to sort all the fields on a model before interrogating them for
        which form field class they need.
        """
        if isinstance(other, Field):
            return self.creation_counter < other.creation_counter
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Field):
            return self.creation_counter == other.creation_counter
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash(self.creation_counter)

    def has_default(self) -> bool:
        """Return a boolean of whether this field has a default value."""
        return self.default is not NOT_PROVIDED

    def get_default(self) -> Any:
        """Return the default value for this field."""
        return self._get_default()

    def check(self, **kwargs) -> List[Union[checks.Error, checks.Warning]]:
        return [
            *self._check_field_name(),
            *self._check_choices(),
            *self._check_null_allowed_for_primary_keys(),
            *self._check_validators(),
        ]

    def _check_field_name(self) -> List[checks.Error]:
        """
        Check if field name is valid, i.e. 1) does not end with an
        underscore, 2) does not contain "__" and 3) is not "pk".
        """
        if cast(str, self.name).endswith('_'):
            return [
                checks.Error(
                    'Field names must not end with an underscore.',
                    obj=self,
                    id='fields.E001',
                )
            ]
        if LOOKUP_SEP in cast(str, self.name):
            return [
                checks.Error(
                    'Field names must not contain "%s".' % (LOOKUP_SEP,),
                    obj=self,
                    id='fields.E002',
                )
            ]
        if self.name == 'pk':
            return [
                checks.Error(
                    "'pk' is a reserved word that cannot be used as a field name.",
                    obj=self,
                    id='fields.E003',
                )
            ]
        return []

    def _check_choices(self) -> List[checks.Error]:
        if not self.choices:
            return []

        def is_value(value):
            return isinstance(value, str)

        if is_value(self.choices):
            return [
                checks.Error(
                    "'choices' must be an iterable (e.g., a list or tuple).",
                    obj=self,
                    id='fields.E004',
                )
            ]

        # Expect [group_name, [value, display]]
        for choices_group in self.choices:
            try:
                group_name, group_choices = choices_group
            except ValueError:
                # Containing non-pairs
                break
            try:
                if not all(
                    is_value(value) and is_value(human_name)
                    for value, human_name in group_choices
                ):
                    break
            except (TypeError, ValueError):
                # No groups, choices in the form [value, display]
                value, human_name = group_name, group_choices
                if not is_value(value) or not is_value(human_name):
                    break

            # Special case: choices=['ab']
            if isinstance(choices_group, str):
                break
        else:
            return []

        return [
            checks.Error(
                "'choices' must be an iterable containing "
                "(actual value, human readable name) tuples.",
                obj=self,
                id='fields.E005',
            )
        ]

    def _check_null_allowed_for_primary_keys(self) -> List[checks.Error]:
        if (self.primary_key and self.null):
            return [
                checks.Error(
                    'Primary keys must not have null=True.',
                    hint=('Set null=False on the field, or '
                          'remove primary_key=True argument.'),
                    obj=self,
                    id='fields.E007',
                )
            ]
        return []

    def _check_validators(self) -> List[checks.Error]:
        errors: List[checks.Error] = []
        for i, validator in enumerate(self.validators):
            if not callable(validator):
                errors.append(
                    checks.Error(
                        "All 'validators' must be callable.",
                        hint=(
                            "validators[{i}] ({repr}) isn't a function or "
                            "instance of a validator class.".format(
                                i=i, repr=repr(validator),
                            )
                        ),
                        obj=self,
                        id='fields.E008',
                    )
                )
        return errors

    @property
    def ldap_attribute(self) -> str:
        return cast(str, self.db_column or self.name)

    @cached_property
    def _get_default(self) -> Callable[[], Any]:
        if self.has_default():
            if callable(self.default):
                return self.default
            return lambda: self.default

        if not self.empty_strings_allowed or self.null:
            return return_None
        return str  # return empty string

    def to_python(self, value: Any) -> Any:
        return value

    @cached_property
    def validators(self) -> List[Validator]:
        return [*self.default_validators, *self._validators]

    def run_validators(self, value: Any) -> None:
        if value in self.empty_values:
            return

        errors = []
        for v in self.validators:
            try:
                v(value)
            except exceptions.ValidationError as e:
                if hasattr(e, 'code') and e.code in self.error_messages:
                    e.message = self.error_messages[e.code]
                errors.extend(e.error_list)

        if errors:
            raise exceptions.ValidationError(errors)

    def validate(self, value: Any, model_instance: "Model") -> None:
        """
        Validate value and raise ValidationError if necessary. Subclasses
        should override this to provide validation logic.
        """
        if not self.editable:
            # Skip validation for non-editable fields.
            return

        if self.choices and value not in self.empty_values:
            for option_key, option_value in self.choices:
                if isinstance(option_value, (list, tuple)):
                    # This is an optgroup, so look inside the group for
                    # options.
                    for optgroup_key, optgroup_value in option_value:  # pylint: disable=unused-variable
                        if value == optgroup_key:
                            return
                elif value == option_key:
                    return
            raise exceptions.ValidationError(
                self.error_messages['invalid_choice'],
                code='invalid_choice',
                params={'value': value},
            )

        if value is None and not self.null:
            raise exceptions.ValidationError(self.error_messages['null'], code='null')

        if not self.blank and value in self.empty_values:
            raise exceptions.ValidationError(self.error_messages['blank'], code='blank')

    def pre_save(self, model_instance: "Model", add: bool) -> Any:
        """Return field's value just before saving."""
        return getattr(model_instance, self.attname)

    def clean(self, value: Any, model_instance: "Model") -> Any:
        """
        Convert the value's type and run validation. Validation errors
        from to_python() and validate() are propagated. Return the correct
        value if no error is raised.
        """
        value = self.to_python(value)
        self.validate(value, model_instance)
        self.run_validators(value)
        return value

    def set_attributes_from_name(self, name: str) -> None:
        self.name = self.name or name
        self.attname = self.name
        if self.verbose_name is None and self.name:
            self.verbose_name = self.name.replace('_', ' ')

    def get_choices(
        self,
        include_blank: bool = True,
        blank_choice: List[Tuple[str, str]] = None,
        limit_choices_to: Any = None
    ):
        """
        Return choices with a default blank choices included, for use as
        <select> choices for this field.
        """
        if not blank_choice:
            blank_choice = BLANK_CHOICE_DASH
        if self.choices:
            choices = list(self.choices)
            if include_blank:
                blank_defined = any(choice in ('', None) for choice, _ in self.flatchoices)
                if not blank_defined:
                    choices = blank_choice + choices
            return choices
        return blank_choice

    def limit_choices_to(self):
        raise NotImplementedError

    def _get_flatchoices(self) -> List[Tuple[str, str]]:
        """Flattened version of choices tuple."""
        flat: List[Tuple[str, str]] = []
        for choice, value in self.choices:
            if isinstance(value, (list, tuple)):
                flat.extend(value)
            else:
                flat.append((choice, value))
        return flat
    flatchoices = property(_get_flatchoices)

    def formfield(
        self,
        form_class: Type[forms.Field] = None,
        choices_form_class: Type[forms.Field] = None,
        **kwargs
    ) -> forms.Field:
        """Return a django.forms.Field instance for this field."""
        defaults: Dict[str, Any] = {
            'required': not self.blank,
            'label': capfirst(self.verbose_name),
            'help_text': self.help_text
        }
        if self.has_default():
            if callable(self.default):
                defaults['initial'] = self.default
                defaults['show_hidden_initial'] = True
            else:
                defaults['initial'] = self.get_default()
        if self.choices:
            # Fields with choices get special treatment.
            include_blank = (self.blank or
                             not (self.has_default() or 'initial' in kwargs))
            defaults['choices'] = self.get_choices(include_blank=include_blank)
            defaults['coerce'] = self.to_python
            if self.null:
                defaults['empty_value'] = None
            if choices_form_class is not None:
                form_class = choices_form_class
            else:
                form_class = forms.TypedChoiceField
            # Many of the subclass-specific formfield arguments (min_value,
            # max_value) don't apply for choice fields, so be sure to only pass
            # the values that TypedChoiceField will understand.
            for k in list(kwargs):
                if k not in ('coerce', 'empty_value', 'choices', 'required',
                             'widget', 'label', 'initial', 'help_text',
                             'error_messages', 'show_hidden_initial', 'disabled'):
                    del kwargs[k]
        defaults.update(kwargs)
        if form_class is None:
            form_class = forms.CharField
        return form_class(**defaults)

    def value_from_object(self, obj: "Model") -> Any:
        return getattr(obj, cast(str, self.name))

    def from_db_value(self, value: List[bytes]) -> Union[List[str], Optional[Any]]:
        """
        Take data for one attribute from LDAP and convert it to our internal
        python format.

        ``value`` will always be a list of byte strings.

        Subclasses should implement the actual logic for this, but
        first call ``super().from_db_value(value)`` to convert the byte
        strings in the list to unicode strings.

        :rtype: varies
        """
        return [b.decode('utf-8') for b in value]

    def to_db_value(self, value: Any) -> Dict[str, List[bytes]]:
        # Subclasses should implement this and do proper casting of the value
        # from our internal data type to the appropriate value to stuff into LDAP
        # and then call super().to_db_value(value)
        if value is None:
            value = []
        if not isinstance(value, list):
            if value not in self.empty_values:
                value = [value]
            else:
                value = []
        # LDAP doesn't like unicode strings; it wants bytes.
        cleaned = []
        for item in value:
            if isinstance(item, str):
                item = item.encode('utf-8')
            cleaned.append(item)
        return {self.ldap_attribute: cleaned}

    def value_to_string(self, obj: "Model") -> str:
        return str(self.value_from_object(obj))

    def save_form_data(self, obj: "Model", value: Any) -> None:
        """
        This is what django.forms.models.ModelForm calls to save data to our
        LdapModel instance.
        """
        setattr(obj, cast(str, self.name), value)

    def contribute_to_class(self, cls, name: str) -> None:
        """
        Register the field with the model class it belongs to.
        """
        self.set_attributes_from_name(name)
        self.model = cls
        self.check()
        cls._meta.add_field(self)
        if self.choices:
            setattr(cls, 'get_{}_display'.format(self.name), partialmethod(cls._get_FIELD_display, field=self))


class BooleanField(Field):

    """
    This is a boolean field which stores data internally as bool() but stores
    the strings 'true' and 'false' in LDAP.
    """

    empty_strings_allowed: bool = False

    default_error_messages: Dict[str, str] = {
        'invalid': _("'%(value)s' value must be either True or False."),
        'invalid_nullable': _("'%(value)s' value must be either True, False, or None."),
    }

    description: str = _("Boolean (Either True or False)")  # type: ignore

    LDAP_TRUE: str = 'true'
    LDAP_FALSE: str = 'false'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.blank = True

    def to_python(self, value: Union[None, bool, str]) -> Optional[bool]:
        if self.null and value in self.empty_values:
            return None
        if value in (True, False):
            # if value is 1 or 0 than it's equal to True or False, but we want
            # to return a true bool for semantic reasons.
            return bool(value)
        if value in ('t', 'True', '1'):
            return True
        if value in ('f', 'False', '0'):
            return False
        raise exceptions.ValidationError(
            self.error_messages['invalid_nullable' if self.null else 'invalid'],
            code='invalid',
            params={'value': value},
        )

    def from_db_value(self, value: List[bytes]) -> Optional[bool]:
        db_value = cast(List[str], super().from_db_value(value))
        if db_value == []:
            return None
        if db_value[0].lower() == self.LDAP_TRUE:
            return True
        if db_value[0].lower() == self.LDAP_FALSE:
            return False
        if db_value == [self.LDAP_TRUE]:
            return True
        if db_value == [self.LDAP_FALSE]:
            return False
        raise ValueError('Field "{}" (BooleanField) on model {} got got unexpected data from LDAP: {}'.format(
            self.name,
            self.model._meta.object_name,  # type: ignore
            db_value
        ))

    def to_db_value(self, value: Optional[bool]) -> Dict[str, List[bytes]]:
        db_value: Optional[str] = None
        if value is not None:
            if value:
                db_value = self.LDAP_TRUE
            else:
                db_value = self.LDAP_FALSE
        return super().to_db_value(db_value)

    def formfield(
        self,
        form_class: Type[forms.Field] = None,
        choices_form_class: Type[forms.Field] = None,
        **kwargs
    ) -> forms.Field:
        if self.choices:
            include_blank = not (self.has_default() or 'initial' in kwargs)
            defaults = {'choices': self.get_choices(include_blank=include_blank)}
        else:
            form_class = forms.NullBooleanField if self.null else forms.BooleanField
            # In HTML checkboxes, 'required' means "must be checked" which is
            # different from the choices case ("must select some value").
            # required=False allows unchecked checkboxes.
            defaults = {
                'form_class': form_class,
                'required': False
            }
        return super().formfield(**{**defaults, **kwargs})


class AllCapsBooleanField(BooleanField):

    LDAP_TRUE: str = 'TRUE'
    LDAP_FALSE: str = 'FALSE'


class CharField(Field):

    description: str = _("String (up to %(max_length)s)")  # type: ignore

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.validators.append(dj_validators.MaxLengthValidator(self.max_length))

    def to_python(self, value: Optional[str]) -> Optional[str]:
        if isinstance(value, str) or value is None:
            return value
        return str(value)

    def from_db_value(self, value: List[bytes]) -> Optional[str]:
        db_value = cast(List[str], super().from_db_value(value))
        if db_value == []:
            return None
        return db_value[0]

    def formfield(
        self,
        form_class: Type[forms.Field] = None,
        choices_form_class: Type[forms.Field] = None,
        **kwargs
    ) -> forms.Field:
        # Passing max_length to forms.CharField means that the value's length
        # will be validated twice. This is considered acceptable since we want
        # the value in the form field (to pass into widget for example).
        defaults: Dict[str, Any] = {
            'form_class': form_class,
            'choices_form_class': choices_form_class,
            'max_length': self.max_length
        }
        if self.null:
            defaults['empty_value'] = None
        defaults.update(kwargs)
        return super().formfield(**defaults)


class DateField(Field):

    empty_strings_allowed: bool = False
    default_error_messages: Dict[str, str] = {
        'invalid': _("'%(value)s' value has an invalid date format. It must be "
                     "in YYYY-MM-DD format."),
        'invalid_date': _("'%(value)s' value has the correct format (YYYY-MM-DD) "
                          "but it is an invalid date."),
    }
    description: str = _("Date (without time)")  # type: ignore

    LDAP_DATETIME_FORMAT: str = "%Y%m%d"

    def __init__(
        self,
        verbose_name: str = None,
        name: str = None,
        auto_now: bool = False,
        auto_now_add: bool = False,
        **kwargs
    ):
        self.auto_now, self.auto_now_add = auto_now, auto_now_add
        if auto_now or auto_now_add:
            kwargs['editable'] = False
            kwargs['blank'] = True
        super().__init__(verbose_name, name, **kwargs)

    def check(self, **kwargs) -> List[Union[checks.Error, checks.Warning]]:
        return [
            *super().check(),
            *self._check_fix_default_value(),
        ]

    def _check_fix_default_value(self) -> List[checks.Warning]:
        """
        Warn that using an actual date or datetime value is probably wrong;
        it's only evaluated on server startup.
        """
        if not self.has_default():
            return []

        now = timezone.now()
        if not timezone.is_naive(now):
            now = timezone.make_naive(now, timezone.utc)
        value = self.default
        if isinstance(value, datetime.datetime):
            if not timezone.is_naive(value):
                value = timezone.make_naive(value, timezone.utc)
            value = value.date()
        elif isinstance(value, datetime.date):
            # Nothing to do, as dates don't have tz information
            pass
        else:
            # No explicit date / datetime value -- no checks necessary
            return []
        offset = datetime.timedelta(days=1)
        lower = (now - offset).date()
        upper = (now + offset).date()
        if lower <= value <= upper:
            return [
                checks.Warning(
                    'Fixed default value provided.',
                    hint='It seems you set a fixed date / time / datetime '
                         'value as default for this field. This may not be '
                         'what you want. If you want to have the current date '
                         'as default, use `django.utils.timezone.now`',
                    obj=self,
                    id='fields.W161',
                )
            ]

        return []

    def to_python(self, value: Optional[Union[str, datetime.date, datetime.datetime]]) -> Optional[datetime.date]:
        if value is None:
            return value
        if isinstance(value, datetime.datetime):
            if settings.USE_TZ and timezone.is_aware(value):
                # Convert aware datetimes to the default time zone
                # before casting them to dates (#17742).
                default_timezone = timezone.get_default_timezone()
                value = timezone.make_naive(value, default_timezone)
            return value.date()
        if isinstance(value, datetime.date):
            return value

        try:
            parsed = parse_date(value)
            if parsed is not None:
                return parsed
        except ValueError:
            raise exceptions.ValidationError(
                self.error_messages['invalid_date'],
                code='invalid_date',
                params={'value': value},
            )

        raise exceptions.ValidationError(
            self.error_messages['invalid'],
            code='invalid',
            params={'value': value},
        )

    def pre_save(self, model_instance: "Model", add: bool) -> Optional[datetime.date]:
        if self.auto_now or (self.auto_now_add and add):
            value = datetime.date.today()
            setattr(model_instance, self.attname, value)
            return value
        return super().pre_save(model_instance, add)

    def contribute_to_class(self, cls, name: str, **kwargs) -> None:
        super().contribute_to_class(cls, name, **kwargs)
        if not self.null:
            setattr(
                cls, 'get_next_by_%s' % self.name,
                partialmethod(cls._get_next_or_previous_by_FIELD, field=self, is_next=True)
            )
            setattr(
                cls, 'get_previous_by_%s' % self.name,
                partialmethod(cls._get_next_or_previous_by_FIELD, field=self, is_next=False)
            )

    def from_db_value(self, value: List[bytes]) -> Optional[datetime.date]:
        db_value = super().from_db_value(value)
        if not db_value:
            return None
        dt = db_value[0]
        ts = datetime.datetime.strptime(dt, self.LDAP_DATETIME_FORMAT)
        return datetime.date(year=ts.year, month=ts.month, day=ts.day)

    def to_db_value(self, value: Optional[Union[datetime.date, datetime.datetime]]) -> Dict[str, List[bytes]]:
        db_value = None
        if value:
            db_value = value.strftime(self.LDAP_DATETIME_FORMAT)
        return super().to_db_value(db_value)

    def value_to_string(self, obj: "Model") -> str:
        val = self.value_from_object(obj)
        return '' if val is None else val.isoformat()

    def formfield(
        self,
        form_class: Type[forms.Field] = None,
        choices_form_class: Type[forms.Field] = None,
        **kwargs
    ) -> forms.Field:
        if not form_class:
            form_class = forms.DateField
        return super().formfield(
            form_class=form_class,
            choices_form_class=choices_form_class,
            **kwargs,
        )


class DateTimeField(DateField):

    LDAP_DATETIME_FORMATS: List[str] = [
        "%Y%m%d%H%M%SZ",
        "%Y%m%d%H%M%S+0000"
    ]
    LDAP_DATETIME_FORMAT: str = "%Y%m%d%H%M%S+0000"

    empty_strings_allowed: bool = False
    default_error_messages: Dict[str, str] = {
        'invalid': _("'%(value)s' value has an invalid format. It must be in "
                     "YYYY-MM-DD HH:MM[:ss[.uuuuuu]][TZ] format."),
        'invalid_date': _("'%(value)s' value has the correct format "
                          "(YYYY-MM-DD) but it is an invalid date."),
        'invalid_datetime': _("'%(value)s' value has the correct format "
                              "(YYYY-MM-DD HH:MM[:ss[.uuuuuu]][TZ]) "
                              "but it is an invalid date/time."),
        'invalid_ldap_datetime': _("LDAP datetime '%(value)s' value is not in a supported format"),
    }
    description: str = _("Date (with time)")

    def _check_fix_default_value(self) -> List[checks.Warning]:
        """
        Warn that using an actual date or datetime value is probably wrong;
        it's only evaluated on server startup.
        """
        if not self.has_default():
            return []

        now = timezone.now()
        if not timezone.is_naive(now):
            now = timezone.make_naive(now, timezone.utc)
        value = self.default
        if isinstance(value, datetime.datetime):
            second_offset = datetime.timedelta(seconds=10)
            lower = now - second_offset
            upper = now + second_offset
            if timezone.is_aware(value):
                value = timezone.make_naive(value, timezone.utc)
        elif isinstance(value, datetime.date):
            second_offset = datetime.timedelta(seconds=10)
            lower = now - second_offset
            lower = datetime.datetime(lower.year, lower.month, lower.day)
            upper = now + second_offset
            upper = datetime.datetime(upper.year, upper.month, upper.day)
            value = datetime.datetime(value.year, value.month, value.day)
        else:
            # No explicit date / datetime value -- no checks necessary
            return []
        if lower <= value <= upper:
            return [
                checks.Warning(
                    'Fixed default value provided.',
                    hint='It seems you set a fixed date / time / datetime '
                         'value as default for this field. This may not be '
                         'what you want. If you want to have the current date '
                         'as default, use `django.utils.timezone.now`',
                    obj=self,
                    id='fields.W161',
                )
            ]

        return []

    def to_python(self, value: Optional[Union[str, datetime.datetime, datetime.date]]) -> Optional[datetime.datetime]:
        if value is None:
            return value
        if isinstance(value, datetime.datetime):
            return value
        if isinstance(value, datetime.date):
            value = datetime.datetime(value.year, value.month, value.day)
            if settings.USE_TZ:
                # For backwards compatibility, interpret naive datetimes in
                # local time. This won't work during DST change, but we can't
                # do much about it, so we let the exceptions percolate up the
                # call stack.
                warnings.warn("DateTimeField %s.%s received a naive datetime "
                              "(%s) while time zone support is active." %
                              (cast(Type["Model"], self.model).__name__, self.name, value),
                              RuntimeWarning)
                default_timezone = timezone.get_default_timezone()
                value = timezone.make_aware(value, default_timezone)
            return value

        try:
            parsed = parse_datetime(value)
            if parsed is not None:
                return parsed
        except ValueError:
            raise exceptions.ValidationError(
                self.error_messages['invalid_datetime'],
                code='invalid_datetime',
                params={'value': value},
            )

        try:
            parsed_date = parse_date(value)
            if parsed_date is not None:
                return datetime.datetime(parsed_date.year, parsed_date.month, parsed_date.day)
        except ValueError:
            raise exceptions.ValidationError(
                self.error_messages['invalid_date'],
                code='invalid_date',
                params={'value': value},
            )

        raise exceptions.ValidationError(
            self.error_messages['invalid'],
            code='invalid',
            params={'value': value},
        )

    def from_db_value(self, value: List[bytes]) -> Optional[datetime.datetime]:
        db_value = Field.from_db_value(self, value)
        if not db_value:
            return None
        dt_str = db_value[0]
        dt: Optional[datetime.datetime] = None
        for fmt in self.LDAP_DATETIME_FORMATS:
            try:
                dt = datetime.datetime.strptime(dt_str, fmt)
            except ValueError:
                pass
            else:
                break
        if not isinstance(dt, datetime.datetime):
            raise exceptions.ValidationError(
                self.error_messages['invalid_ldap_datetime'],
                code='invalid_ldap_datetime',
                params={'value': value},
            )
        dt = pytz.utc.localize(dt)
        return dt

    def to_db_value(self, value: Optional[Union[datetime.datetime, datetime.date]]) -> Dict[str, List[bytes]]:
        dt_str = None
        if value:
            utc = pytz.utc
            dt_str = cast(datetime.datetime, value).astimezone(utc).strftime(self.LDAP_DATETIME_FORMAT)
        return Field.to_db_value(self, dt_str)

    def pre_save(self, model_instance: "Model", add: bool) -> Any:
        if self.auto_now or (self.auto_now_add and add):
            value = timezone.now()
            setattr(model_instance, self.attname, value)
            return value
        return super().pre_save(model_instance, add)

    def formfield(
        self,
        form_class: Type[forms.Field] = None,
        choices_form_class: Type[forms.Field] = None,
        **kwargs
    ) -> forms.Field:
        if not form_class:
            form_class = forms.DateTimeField
        return super().formfield(
            form_class=form_class,
            choices_form_class=choices_form_class,
            **kwargs,
        )


class EmailField(CharField):
    default_validators: List[Validator] = [dj_validators.validate_email]
    description: str = _("Email address")

    def __init__(self, *args, **kwargs):
        # max_length=254 to be compliant with RFCs 3696 and 5321
        kwargs.setdefault('max_length', 254)
        super().__init__(*args, **kwargs)

    def formfield(
        self,
        form_class: Type[forms.Field] = None,
        choices_form_class: Type[forms.Field] = None,
        **kwargs
    ) -> forms.Field:
        # As with CharField, this will cause email validation to be performed
        # twice.
        if not form_class:
            form_class = forms.EmailField
        return super().formfield(
            form_class=form_class,
            choices_form_class=choices_form_class,
            **kwargs,
        )


class EmailForwardField(CharField):
    default_validators: List[Validator] = [validate_email_forward]
    description: str = _("Email address")


class IntegerField(Field):
    empty_strings_allowed: bool = False
    default_error_messages: Dict[str, str] = {
        'invalid': _("'%(value)s' value must be an integer."),
    }
    description: str = _("Integer")  # type: ignore

    def check(self, **kwargs) -> List[Union[checks.Error, checks.Warning]]:
        return [
            *super().check(),
            *self._check_max_length_warning(),
        ]

    def _check_max_length_warning(self) -> List[checks.Warning]:
        if self.max_length is not None:
            return [
                checks.Warning(
                    "'max_length' is ignored when used with IntegerField",
                    hint="Remove 'max_length' from field",
                    obj=self,
                    id='fields.W122',
                )
            ]
        return []

    def from_db_value(self, value: List[bytes]) -> Optional[int]:
        db_value = super().from_db_value(value)
        if not db_value:
            return None
        return self.to_python(db_value[0])

    def to_db_value(self, value: Optional[int]) -> Dict[str, List[bytes]]:
        db_value: Optional[str] = str(value) if value else None
        return super().to_db_value(db_value)

    def to_python(self, value: str) -> int:
        if value is None:
            return value
        try:
            return int(value)
        except (TypeError, ValueError):
            raise exceptions.ValidationError(
                self.error_messages['invalid'],
                code='invalid',
                params={'value': value},
            )

    def formfield(
        self,
        form_class: Type[forms.Field] = None,
        choices_form_class: Type[forms.Field] = None,
        **kwargs
    ) -> forms.Field:
        if not form_class:
            form_class = forms.IntegerField
        return super().formfield(
            form_class=form_class,
            choices_form_class=choices_form_class,
            **kwargs,
        )


class CharListField(CharField):
    description: str = _("List of strings (each up to %(max_length)s)")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.empty_values += ('[]', )
        self.validators.append(dj_validators.MaxLengthValidator(self.max_length))

    def get_default(self) -> List[str]:
        if self.default is None:
            return []
        return self._get_default()

    def from_db_value(self, value):
        """
        .. note::

            This is important because we don't want CharField.from_db_value() to
            execute; we only want Field.from_db_value().  CharField.from_db_value()
            turns the list that Field.from_db_value() returns into a string.  We
            actually want the list.
        """
        return Field.from_db_value(self, value)

    def to_python(self, value: Optional[Union[str, List[str]]]) -> List[str]:  # type: ignore
        if not value:
            return []
        if isinstance(value, list):
            return value
        return value.splitlines()

    def formfield(
        self,
        form_class: Type[forms.Field] = None,
        choices_form_class: Type[forms.Field] = None,
        **kwargs
    ) -> forms.Field:
        # Passing max_length to forms.CharField means that the value's length
        # will be validated twice. This is considered acceptable since we want
        # the value in the form field (to pass into widget for example).
        defaults: Dict[str, Any] = {
            'max_length': self.max_length,
            'form_class': ldap_forms.CharListField,
        }
        if self.null:
            defaults['empty_value'] = []
        defaults.update(kwargs)
        return super().formfield(**defaults)


class CaseInsensitiveSHA1Field(CharField):
    """
    This is a readonly field that stores its value as a lowercased SHA1 hash.
    Use this when you want to store a secret value that you will only ever
    compare to a similar hashed value.
    """

    @staticmethod
    def hash_value(value: Optional[str]) -> Optional[str]:
        if value:
            return hashlib.sha1(value.lower().encode('utf-8')).hexdigest()
        return None

    def to_db_value(self, value: str) -> Dict[str, List[bytes]]:
        return super().to_db_value(self.hash_value(value))


class PasswordField(CharField):

    def hash_password(self, password: str) -> bytes:
        return password.encode('utf-8')

    def to_db_value(self, value: str) -> Dict[str, List[bytes]]:
        hashed_password = self.hash_password(value)
        return super().to_db_value(hashed_password)


class LDAPPasswordField(PasswordField):

    def hash_password(self, password: str) -> bytes:
        salt = os.urandom(8)
        h = hashlib.sha1(password.encode('utf-8'))
        h.update(salt)
        pwhash = "{SSHA}".encode('utf-8') + encode(h.digest() + salt)
        return pwhash


class ADPasswordField(PasswordField):

    def hash_password(self, password: str) -> bytes:
        return "\"{}\"".format(password).encode("utf-16-le")



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='django-ldaporm',
    version='1.0.16',
    description='A Django ORM-like interface for ldap objectclasses',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['django', 'ldap'],
    author="Caltech IMSS ADS",
    author_email="imss-ads-staff@caltech.edu",
    url='https://github.com/caltechads/django-ldaporm',
    packages=find_packages(exclude=['bin']),
    include_package_data=True,
    package_data={'ldaporm': ["py.typed"]},
    install_requires=[
        'pytz',
        'ldap_filter',
        'python-ldap',
    ],
    classifiers=[
        "Programming Language :: Python :: 3"
    ],
)
